"""Utilities for working with TMDb models."""


async def overlapping_movies(people, client=None):
    """Find movies that the same people have been in.

    Arguments:
      people (:py:class:`collections.abc.Sequence`): The
        :py:class:`~.Person` objects to find overlapping movies for.
      client (:py:class:`~.TMDbClient`, optional): The TMDb client
        to extract additional information about the overlap.

    Returns:
      :py:class:`list`: The relevant :py:class:`~.Movie` objects.

    """
    return await _overlap(people, 'movie_credits', client, 'get_movie')


async def overlapping_actors(movies, client=None):
    """Find actors that appear in the same movies.

    Arguments:
      movies (:py:class:`collections.abc.Sequence`): The
        :py:class:`~.Movie` objects to find overlapping actors for.
      client (:py:class:`~.TMDbClient`, optional): The TMDb client
        to extract additional information about the overlap.

    Returns:
      :py:class:`list`: The relevant :py:class:`~.Person` objects.

    """
    return await _overlap(movies, 'cast', client, 'get_person')

async def find_overlapping_movies(names, client):
    """Find movies that the same people have been in.

    Warning:
      This function requires two API calls per name submitted, plus
      one API call per overlapping movie in the result; it is therefore
      relatively slow.

    Arguments:
      names (:py:class:`collections.abc.Sequence`): The names of the
        people to find overlapping movies for.
      client (:py:class:`~.TMDbClient`): The TMDb client.

    Returns:
      :py:class:`list`: The relevant :py:class:`~.Movie` objects.

    """
    return await _find_overlap(names, client, 'find_person', 'get_person',
                               overlapping_movies)


async def find_overlapping_actors(titles, client):
    """Find actors that have been in the same movies.

    Warning:
      This function requires two API calls per title submitted, plus
      one API call per overlapping person in the result; it is therefore
      relatively slow.

    Arguments:
      titles (:py:class:`collections.abc.Sequence`): The titles of the
        movies to find overlapping actors for.
      client (:py:class:`~.TMDbClient`): The TMDb client.

    Returns:
      :py:class:`list`: The relevant :py:class:`~.Person` objects.

    """
    return await _find_overlap(titles, client, 'find_movie', 'get_movie',
                               overlapping_actors)


async def _overlap(items, overlap_attr, client=None, get_method=None):
    """Generic overlap implementation.

    Arguments:
      item (:py:class:`collections.abc.Sequence`): The objects to
        find overlaps for.
      overlap_attr (:py:class:`str`): The attribute of the items to use
        as input for the overlap.
      client (:py:class:`~.TMDbClient`, optional): The TMDb client
        to extract additional information about the overlap.
      get_method (:py:class:`str`, optional): The method of the
        client to use for extracting additional information.

    Returns:
      :py:class:`list`: The relevant result objects.

    """
    overlap = set.intersection(*(getattr(item, overlap_attr) for item in items))
    if client is None or get_method is None:
        return overlap
    results = []
    for item in overlap:
        result = await getattr(client, get_method)(id_=item.id_)
        results.append(result)
    return results


async def _find_overlap(queries, client, find_method, get_method,
                        overlap_function):
    """Generic find and overlap implementation.

    Arguments
      names (:py:class:`collections.abc.Sequence`): The queries of the
        people to find overlaps for.
      client (:py:class:`~.TMDbClient`): The TMDb client.
      find_method (:py:class:`str`): The name of the client method to
        use for finding candidates.
      get_method (:py:class:`str`): The name of the client method to
        use for getting detailed information on a candidate.
      overlap_function (:py:class:`collections.abc.Callable`): The
        function to call for the resulting overlap.



    """
    results = []
    for query in queries:
        candidates = await getattr(client, find_method)(query)
        if not candidates:
            raise ValueError('no result found for {!r}'.format(query))
        result = await getattr(client, get_method)(id_=candidates[0].id_)
        results.append(result)
    return await overlap_function(results, client)



"""Core API wrapper functionality, adapted from `Flash Services`_.

.. _Flash Services:
    https://pypi.python.org/pypi/flash-services

"""
# pylint: disable=too-few-public-methods
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from datetime import datetime, timezone
from urllib.parse import urlencode

from dateutil.parser import parse


class Service(metaclass=ABCMeta):
    """Abstract base class for API wrapper services."""

    REQUIRED = set()
    """:py:class:`set`: The service's required configuration keys."""

    ROOT = ''
    """:py:class:`str`: The root URL for the API."""

    @abstractmethod
    def __init__(self, *_, **kwargs):
        self.service_name = kwargs.get('name')

    @property
    def headers(self):
        """Get the headers for the service requests.

        Returns:
          :py:class:`dict`: The header mapping.

        """
        return {}

    def url_builder(self, endpoint, *, root=None, params=None, url_params=None):
        """Create a URL for the specified endpoint.

        Arguments:
          endpoint (:py:class:`str`): The API endpoint to access.
          root: (:py:class:`str`, optional): The root URL for the
            service API.
          params: (:py:class:`dict`, optional): The values for format
            into the created URL (defaults to ``None``).
          url_params: (:py:class:`dict`, optional): Parameters to add
            to the end of the URL (defaults to ``None``).

        Returns:
          :py:class:`str`: The resulting URL.

        """
        if root is None:
            root = self.ROOT
        return ''.join([
            root,
            endpoint,
            '?' + urlencode(url_params) if url_params else '',
        ]).format(**params or {})

    @staticmethod
    def calculate_timeout(headers):
        """Extract request timeout from ``Retry-After`` header.

        Notes:
          Per :rfc:`2616#section-14.37`, the ``Retry-After`` header can
          be either an integer number of seconds or an HTTP date. This
          function can handle either.

        Arguments:
          headers (:py:class:`aiohttp.CIMultiDictProxy`): The response
            headers.

        Returns:
          :py:class:`int`: The timeout, in seconds.

        """
        after = headers['Retry-After']
        try:
            return int(after)
        except ValueError:
            date_after = parse(after)
        utc_now = datetime.now(tz=timezone.utc)
        return int((date_after - utc_now).total_seconds())


class TokenAuthMixin:
    """Mix-in class for implementing token authentication.

    Arguments:
      api_token (:py:class:`str`): A valid API token.

    """

    def __init__(self, *, api_token, **kwargs):
        self.api_token = api_token
        super().__init__(**kwargs)


class UrlParamMixin(TokenAuthMixin):
    """Mix-in class for implementing URL parameter authentication.

    Attributes:
      AUTH_PARAM (:py:class:`str`): The name of the URL parameter to
        supply the token as.
    """

    def url_builder(self, endpoint, params=None, url_params=None):
        """Add authentication URL parameter."""
        if url_params is None:
            url_params = OrderedDict()
        url_params[self.AUTH_PARAM] = self.api_token
        return super().url_builder(
            endpoint,
            params=params,
            url_params=url_params,
        )



"""API client wrapper."""
import asyncio
from collections import OrderedDict
from datetime import datetime, timedelta
from http import HTTPStatus
import json
import logging

import aiohttp

from .core import UrlParamMixin, Service
from .models import Movie, Person

logger = logging.getLogger(__name__)


class TMDbClient(UrlParamMixin, Service):
    """Simple wrapper for the `TMDb`_ API.

    .. _TMDb: https://www.themoviedb.org/

    """

    AUTH_PARAM = 'api_key'

    ROOT = 'https://api.themoviedb.org/3/'

    def __init__(self, *, api_token, **kwargs):
        super().__init__(api_token=api_token, **kwargs)
        self.config = dict(data=None, last_update=None)

    @property
    def headers(self):
        return dict(Accept='application/json', **super().headers)

    @property
    def config_expired(self):
        """Whether the configuration data has expired."""
        return (self.config['last_update'] + timedelta(days=2)) < datetime.now()

    async def _update_config(self):
        """Update configuration data if required.

        Notes:
          Per `the documentation`_, this updates the API configuration
          data *"every few days"*.

        .. _the documentation:
          http://docs.themoviedb.apiary.io/#reference/configuration

        """
        if self.config['data'] is None or self.config_expired:
            data = await self.get_data(self.url_builder('configuration'))
            self.config = dict(data=data, last_update=datetime.now())

    async def get_data(self, url):
        """Get data from the TMDb API via :py:func:`aiohttp.get`.

        Notes:
          Updates configuration (if required) on successful requests.

        Arguments:
          url (:py:class:`str`): The endpoint URL and params.

        Returns:
          :py:class:`dict`: The parsed JSON result.

        """
        logger.debug('making request to %r', url)
        with aiohttp.ClientSession() as session:
            async with session.get(url, headers=self.headers) as response:
                body = json.loads((await response.read()).decode('utf-8'))
                if response.status == HTTPStatus.OK:
                    if url != self.url_builder('configuration'):
                        await self._update_config()
                    return body
                elif response.status == HTTPStatus.TOO_MANY_REQUESTS:
                    timeout = self.calculate_timeout(response.headers)
                    logger.warning(
                        'Request limit exceeded, waiting %s seconds',
                        timeout,
                    )
                    await asyncio.sleep(timeout)
                    return await self.get_data(url)
                logger.warning(
                    'request failed %s: %r',
                    response.status,
                    body.get('status_message', '<no message>')
                )

    async def find_movie(self, query):
        """Retrieve movie data by search query.

        Arguments:
          query (:py:class:`str`): Query to search for.

        Returns:
          :py:class:`list`: Possible matches.

        """
        params = OrderedDict([
            ('query', query), ('include_adult', False),
        ])
        url = self.url_builder('search/movie', {}, params)
        data = await self.get_data(url)
        if data is None:
            return
        return [
            Movie.from_json(item, self.config['data'].get('images'))
            for item in data.get('results', [])
        ]

    async def find_person(self, query):
        """Retrieve person data by search query.

        Arguments:
          query (:py:class:`str`): Query to search for.

        Returns:
          :py:class:`list`: Possible matches.

        """
        url = self.url_builder(
            'search/person',
            dict(),
            url_params=OrderedDict([
                ('query', query), ('include_adult', False)
            ]),
        )
        data = await self.get_data(url)
        if data is None:
            return
        return [
            Person.from_json(item, self.config['data'].get('images'))
            for item in data.get('results', [])
        ]

    async def get_movie(self, id_):
        """Retrieve movie data by ID.

        Arguments:
          id_ (:py:class:`int`): The movie's TMDb ID.

        Returns:
          :py:class:`~.Movie`: The requested movie.

        """
        url = self.url_builder(
            'movie/{movie_id}',
            dict(movie_id=id_),
            url_params=OrderedDict(append_to_response='credits'),
        )
        data = await self.get_data(url)
        if data is None:
            return
        return Movie.from_json(data, self.config['data'].get('images'))

    async def get_person(self, id_):
        """Retrieve person data by ID.

        Arguments:
          id_ (:py:class:`int`): The person's TMDb ID.

        Returns:
          :py:class:`~.Person`: The requested person.

        """
        url = self.url_builder(
            'person/{person_id}',
            dict(person_id=id_),
            url_params=OrderedDict(append_to_response='movie_credits'),
        )
        data = await self.get_data(url)
        if data is None:
            return
        return Person.from_json(data, self.config['data'].get('images'))



"""Models representing TMDb resources."""
import logging
from textwrap import dedent

logger = logging.getLogger(__name__)


class BaseModel:
    """Base TMDb model functionality.

    Arguments:
      id_ (:py:class:`int`): The TMDb ID of the object.
      image_config (:py:class:`dict`): The API image configuration.

    """

    CONTAINS = None
    """:py:class:`dict`: Rules for what the model contains."""

    IMAGE_TYPE = None
    """:py:class:`str`: The type of image to use."""

    JSON_MAPPING = dict(id_='id')
    """:py:class:`dict`: The mapping between JSON keys and attributes."""

    image_config = None

    def __init__(self, *, id_, image_path=None, **_):
        self.id_ = id_
        self.image_path = image_path

    def __contains__(self, item):
        if self.CONTAINS is None:
            return False
        attr = self.CONTAINS['attr']
        subclasses = {obj.__name__: obj for obj in
                      BaseModel.__subclasses__()}  # pylint: disable=no-member
        cls = subclasses[self.CONTAINS['type']]
        return isinstance(item, cls) and item in getattr(self, attr)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.id_ == other.id_

    def __hash__(self):
        return hash(self.id_)

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join(['{}={!r}'.format(attr, getattr(self, attr))
                       for attr in self.JSON_MAPPING]),
        )

    @property
    def image_url(self):
        """The fully-qualified image URL."""
        return self._create_image_url(self.image_path, self.IMAGE_TYPE, 200)

    def _create_image_url(self, file_path, type_, target_size):
        """The the closest available size for specified image type.

        Arguments:
          file_path (:py:class:`str`): The image file path.
          type_ (:py:class:`str`): The type of image to create a URL
            for, (``'poster'`` or ``'profile'``).
          target_size (:py:class:`int`): The size of image to aim for (used
            as either width or height).

        """
        if self.image_config is None:
            logger.warning('no image configuration available')
            return
        return ''.join([
            self.image_config['secure_base_url'],
            self._image_size(self.image_config, type_, target_size),
            file_path,
        ])

    @classmethod
    def from_json(cls, json, image_config=None):
        """Create a model instance

        Arguments:
          json (:py:class:`dict`): The parsed JSON data.
          image_config (:py:class:`dict`): The API image configuration
            data.

        Returns:
          :py:class:`BaseModel`: The model instance.

        """
        cls.image_config = image_config
        return cls(**{
            attr: json.get(key) for attr, key in cls.JSON_MAPPING.items()
        })

    @staticmethod
    def _image_size(image_config, type_, target_size):
        """Find the closest available size for specified image type.

        Arguments:
          image_config (:py:class:`dict`): The image config data.
          type_ (:py:class:`str`): The type of image to create a URL
            for, (``'poster'`` or ``'profile'``).
          target_size (:py:class:`int`): The size of image to aim for (used
            as either width or height).

        """
        return min(
            image_config['{}_sizes'.format(type_)],
            key=lambda size: (abs(target_size - int(size[1:]))
                              if size.startswith('w') or size.startswith('h')
                              else 999),
        )


class Movie(BaseModel):
    """Represents a movie.

    Arguments:
      title (:py:class:`str`): The title of the movie.
      cast (:py:class:`set`, optional): The movie's cast.
      synopsis (:py:class:`str`, optional): A synopsis of the movie.

    """

    CONTAINS = dict(attr='cast', image_path='poster_path', type='Person')

    IMAGE_TYPE = 'poster'

    JSON_MAPPING = dict(
        cast='cast',
        image_path='{}_path'.format(IMAGE_TYPE),
        synopsis='overview',
        title='original_title',
        **BaseModel.JSON_MAPPING,
    )

    def __init__(self, *, title, cast=None, poster=None, synopsis=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.cast = cast
        self.poster = poster
        self.synopsis = synopsis
        self.title = title

    def __str__(self):
        if self.synopsis is None:
            return "{0.title} [{0.url}]".format(self)
        return dedent("""
        *{0.title}*

        {0.synopsis}

        For more information see: {0.url}
        """).strip().format(self)

    @property
    def url(self):
        return 'https://www.themoviedb.org/movie/{}'.format(self.id_)

    @classmethod
    def from_json(cls, json, image_config=None):
        json['cast'] = {
            Person.from_json(person) for person in
            json.get('credits', {}).get('cast', [])
        } or None
        return super().from_json(json, image_config)


class Person(BaseModel):
    """Represents a person.

    Arguments:
      name (:py:class:`str`): The person's name.
      movie_credits (:py:class:`set`, optional): The person's movie
        credits.
      biography (:py:class:`str`, optional): A synopsis of the movie.

    """

    CONTAINS = dict(attr='movie_credits', type='Movie')

    IMAGE_TYPE = 'profile'

    JSON_MAPPING = dict(
        biography='biography',
        image_path='{}_path'.format(IMAGE_TYPE),
        movie_credits='movie_credits',
        name='name',
        **BaseModel.JSON_MAPPING,
    )

    def __init__(self, name, biography=None, movie_credits=None, profile=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.biography = biography
        self.movie_credits = movie_credits
        self.name = name
        self.profile = profile

    def __str__(self):
        if self.biography is None:
            return "{0.name} [{0.url}]".format(self)
        return dedent("""
        *{0.name}*

        {0.biography}

        For more information see: {0.url}
        """).strip().format(self)

    @property
    def url(self):
        return 'https://www.themoviedb.org/person/{}'.format(self.id_)

    @classmethod
    def from_json(cls, json, image_config=None):
        json['movie_credits'] = {
            Movie.from_json(movie) for movie in
            json.get('movie_credits', {}).get('cast', [])
        } or None
        return super().from_json(json, image_config)



import io
import os
import re
from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys

PKG_NAME = 'atmdb'

HERE = os.path.abspath(os.path.dirname(__file__))

PATTERN = r'^{target}\s*=\s*([\'"])(.+)\1$'

AUTHOR = re.compile(PATTERN.format(target='__author__'), re.M)
DOCSTRING = re.compile(r'^([\'"])\1\1(.+)\1\1\1$', re.M)
VERSION = re.compile(PATTERN.format(target='__version__'), re.M)


def parse_init():
    with open(os.path.join(HERE, PKG_NAME, '__init__.py')) as f:
        file_data = f.read()
    return [regex.search(file_data).group(2) for regex in
            (AUTHOR, DOCSTRING, VERSION)]


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


long_description = read('README.rst')
author, description, version = parse_init()


class PyTest(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--pylint',
            '--pylint-error-types=FEW',
            '--runslow',
        ]
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)

setup(
    author=author,
    author_email='mail@jonrshar.pe',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3 :: Only',
    ],
    cmdclass={'test': PyTest},
    description=description,
    install_requires=['aiohttp'],
    license='License :: OSI Approved :: ISC License (ISCL)',
    long_description=long_description,
    name=PKG_NAME,
    packages=[PKG_NAME],
    platforms='any',
    tests_require=[
        'asynctest',
        'pylint',
        'pytest',
        'pytest-asyncio',
        'pytest-pylint',
    ],
    url='http://github.com/textbook/atmdb/',
    version=version,
)